// Selective color based blur
// gimp plugin
// version 0.4.0 - development version
//
// to compile & install (on linux):
// for/as current user:
//  $gimptool-2.0 --install colorblur.c
// and for system-wide installation (run as root):
//  $gimptool-2.0 --install-admin colorblur.c
// following might be needed beforehand (on Arch Linux f.e.): 
//  $export LDFLAGS="$LDFLAGS -lm"
//
//contact: tiborb95 at gmail dot com, any feedback welcomed
// or discuss on http://registry.gimp.org/node/25801



#include <libgimp/gimp.h>
#include <stdio.h>
#include <libgimp/gimpui.h>
#include <string.h> //- not needed probably
#include <stdlib.h> // for exit
#include <math.h>


#define VERBOSE FALSE
#define DEBUG FALSE
#define RESPONSE_RESET   99
#define RESPONSE_OK      1
#define TO_FLOAT_W_GAMMA(x) pow((gfloat)x/255,1/2.2)
#define TO_FLOAT(x) (gfloat)x/255
#define TO_INT_W_GAMMA(x) MAX(MIN(ROUND(pow(x,2.2)*255),255),0)
#define TO_INT(x) MAX(MIN(x*255,255),0)
#define SQ(x) x*x
#define ITERATE(x,start,end) for(x=start;x<end;x++)
#define RADIUSDEFAULT 5
#define TRESHDEFAULT 0.02
#define BRTRESHDEF 0.06
#define ITERDEFAULT 6
#define THREADS 5     
#define COLORSDEF 6
#define LINESENSDEF 1.0
//processing modes
#define NOCHANGE 0
#define CARTREDBR 1
#define CARTOONIZE 2
#define DOLINES 3
#define MODEDEF 2     //default
////saturation modes
//#define NOTCHANGESAT 0
//#define DECENTSAT 1
//#define SATANIMEWEAK 2
//#define SATANIME 3
//#define SMODEDEF 0
#define MINDIRCOUNT 2


//structures
typedef struct {
	gfloat tresh;
	gint radius;
	gint iterations;
	gint mode;
	gfloat brtresh;
	gfloat linesens;
	gfloat slidersat;
	gboolean exportlayer;
	gboolean doclean;
	gfloat darkpoint;
	gfloat lightpoint;
	gint mindircount;
	gboolean preview;
	gint channels;
	gint directions[8];
	gfloat treshsq;
	gint shortrad;
	gfloat localweight;
	gint width;
	gint height;
	gfloat breakpoint;
} MyGenVals;
static MyGenVals maindata = {TRESHDEFAULT,RADIUSDEFAULT, ITERDEFAULT,MODEDEF,BRTRESHDEF,LINESENSDEF,1,FALSE,FALSE,0,1,MINDIRCOUNT};

typedef struct {
	gint startline;
	gint endline;
	gint id;
	gint iteration;
	gfloat* rsource;
	gfloat* gsource;
	gfloat* bsource;
	gfloat* rtarget;
	gfloat* gtarget;
	gfloat* btarget;
	gfloat* brsource;
	gfloat* brtarget;
	//gboolean doclean;
	gint stat[3];
} threaddatastr;

typedef struct {
	gboolean good;
	gint len_good;
	gfloat sumbrdist;
	gfloat sumhuedist;
	gfloat sumr;
	gfloat sumb;
	gfloat sumg;
	gfloat sumbr;
} dirstruct;

//variables
static gboolean previewbool;
static guchar *rect_in,*rect_out;
static gboolean *rect_good; // TRUE for pixels that were blurred in good way
static gfloat *r1,*g1,*b1,*r2,*g2,*b2,*br1,*br2; // working layers to be switched
static gfloat *r_orig,*g_orig,*b_orig,*brorig;   //original layer (can be changed!)
static gfloat *br_unch;     // original br data - should not be changed
static gint basepos_diff[8]; //to speed up calculating local basepos
static gint near_pixels[16]={0,1,1,1,1,0,1,-1,0,-1,-1,-1,-1,0,-1,1};

//static gboolean preview;
static gboolean process_image = FALSE;
gint32 image_ID;

//GUI variables
gchar *title = "Selective Color Blurring (0.4.0)";
static GtkWidget *radius_spin,*tresh_spin, *iter_spin, *brtresh_spin,*linesens_spin,*slider_sat;
static GtkWidget *combo;
gchar *mode0 = "Blur the color (only)";
gchar *mode1 = "Cartonize (shrink brightness)";
gchar *mode2 = "Cartoonize   ";
gchar *mode3 = "Create lines ";
gchar *smode0 = "Not saturating      ";
gchar *smode1 = "Saturation - decent  ";
gchar *smode2 = "Saturation - anime  ";
gchar *smode3 = "Saturation - strong  ";
static GtkWidget *cleanbutton;

//multithreading related
static GThread *gth[THREADS];
static threaddatastr thread_data[THREADS];

//functions
static void query (void);
static void run   (const gchar      *name,
                   gint              nparams,
                   const GimpParam  *param,
                   gint             *nreturn_vals,
                   GimpParam       **return_vals);
static void process  (GimpDrawable *drawable,GimpPreview *preview);
GimpPlugInInfo PLUG_IN_INFO = { NULL, NULL, query, run };
static gboolean plugin_gui (GimpDrawable *drawable);
static void response_callback (GtkWidget *widget,  gint response_id);
static void blur_data();
void *blur_step (void *arg);
void *create_lines (void *arg);
static void fill_rect_out();
static void fill_tmp();
static gfloat	gammed_values[256];
inline gfloat get_bright (gfloat R,gfloat G,gfloat B);
static void  mode_changed( GtkComboBox *combo);
void cb_changed(GtkToggleButton *cleanbutton);
void ds_changed(GtkToggleButton *dsbutton);
void do_lines();
inline gint basepos(gint x, gint y,gint ch, gint width);

MAIN()


static void query (void) {
	static GimpParamDef args[] =   {
  		{ GIMP_PDB_INT32, "run-mode", "Run mode" },
    	{ GIMP_PDB_IMAGE, "image", "Input image"  },
    	{ GIMP_PDB_DRAWABLE,"drawable", "Input drawable" },
     	{ GIMP_PDB_INT32,"mode", "Processing mode (0,1,2,3)" },
    	{ GIMP_PDB_INT32,"radius", "radius" },
    	{ GIMP_PDB_INT32,"iterations", "Iterations" },
    	{ GIMP_PDB_FLOAT,"coltresh", "Color treshold" },
    	{ GIMP_PDB_FLOAT,"brtresh", "Brightness treshold" },
    	{ GIMP_PDB_FLOAT,"linesens", "Sensitivity for lines creation" },
    	{ GIMP_PDB_INT32,"docleanup", "Ignore small details (1=True)" },
    	{ GIMP_PDB_FLOAT,"darkpoint", "Modification of 0 brightness point" },
    	{ GIMP_PDB_FLOAT,"lightpoint", "Modification of 0 brightness point" },    	
    	{ GIMP_PDB_FLOAT,"slidersat", "Saturation" }
    	   };

	gimp_install_procedure (
    	"plug-in-colorblur",
    	title,
    	"Selective Color-based Blurring - blurs color information or whole image, intended for improvement of colors or cartoonization.",
    	"Tibor Bamhor",
    	"GPL v.3",
    	"2011",
    	"_Color Selective Blur",
    	"RGB*",
    	GIMP_PLUGIN,
    	G_N_ELEMENTS (args), 0,
    	args, NULL);

	gimp_plugin_menu_register ("plug-in-colorblur",                        
    	"<Image>/Filters/Blur");
}

void static parse_bool(gint value, gint* cur_value,gint position) {
	if (value==-1) return;			
 	if (value>=0 && value <=1) *cur_value=value;
 	else {printf ("  Incorrect argument %1d: %1d (expecting bool - 0/1)\n\n",position, value);exit(1);}
	} 
 	
void static parse_range_int(gint value, gint* cur_value,gint position,gint min, gint max){
	if (value==-1) return;
 	if (value>=min && value <=max) *cur_value=value;
 	else {printf ("  Incorrect argument %1d: %1d (expecting int in range %1d - %1d)\n\n",position, value, min,max);exit(1);}
	}	

void static parse_range_float(gfloat value, gfloat* cur_value,gint position,gfloat min, gfloat max){
	if (value==-1) return;
 	if (value>=min && value <=max) *cur_value=value;
 	else {printf ("  Incorrect argument %1d: %.3f (expecting float in range %.3f - %.3f)\n\n",position, value, min,max);exit(1);}
	}	

static void run (const gchar      *name,
     			gint              nparams,
     			const GimpParam  *param,
     			gint             *nreturn_vals,
     			GimpParam       **return_vals){
	static GimpParam  values[1];
	GimpPDBStatusType status = GIMP_PDB_SUCCESS;
	GimpRunMode       run_mode;
	GimpDrawable     *drawable;

	/* Setting mandatory output values */
	*nreturn_vals = 1;
	*return_vals  = values;

	values[0].type = GIMP_PDB_STATUS;
	values[0].data.d_status = status;
	
	image_ID = param[1].data.d_image; //<- zistujem imageID

	// Getting run_mode - we won't display a dialog if
	// we are in NONINTERACTIVE mode
  	run_mode = param[0].data.d_int32;

  	/*  Get the specified drawable  */
	drawable = gimp_drawable_get (param[2].data.d_drawable);

    switch (run_mode) {
    	case GIMP_RUN_INTERACTIVE:
            /* Get options last values if needed */
            //gimp_get_data ("plug-in-colorblur", &satvals);

            /* Display the dialog */
            if (! plugin_gui (drawable)) return;
            break;

		case GIMP_RUN_NONINTERACTIVE:
			if (nparams != 13) {
				printf (" NON-INTERACTIVE MODE ERROR: wrong number of parameters (%1d, expected: 13). Quitting...\n\n",nparams);
				exit(1);}
				
			if (VERBOSE) printf (" Running non-interactive mode; parameters: %1d\n",nparams);
			parse_range_int(param[3].data.d_int32,&maindata.mode,1,0,3);
			parse_range_int(param[4].data.d_int32,&maindata.radius,2,0,300);
			parse_range_int(param[5].data.d_int32,&maindata.iterations,3,0,300);
			parse_range_float(param[6].data.d_float,&maindata.tresh,4,0,2);
			parse_range_float(param[7].data.d_float,&maindata.brtresh,5,0,1);			
			parse_range_float(param[8].data.d_float,&maindata.linesens,6,0.1,20);				
			parse_bool(param[9].data.d_int32,&maindata.doclean,7);
			parse_range_float(param[10].data.d_float,&maindata.darkpoint,8,-0.5,1.0);	
			parse_range_float(param[11].data.d_float,&maindata.lightpoint,9,0,1.5);	
			parse_range_float(param[12].data.d_float,&maindata.slidersat,10,0,2);	
			 
            //if (nparams != 4)  status = GIMP_PDB_CALLING_ERROR;
            break;

		//case GIMP_RUN_WITH_LAST_VALS:
            ///*  Get options last values if needed  */
            //gimp_get_data ("plug-in-colorblur", &satvals);
            //break;

		default:
            break;
        }

	process (drawable,NULL); 


	gimp_displays_flush ();
	gimp_drawable_detach (drawable);
	
	fflush (stdout );
	
	return;
}

void export_to_layer()
{
	GimpDrawable *new_layer;
	gint layer_ID,basepos_4,basepos_ch;
	gint x,y;
	//gfloat br; //brightness
	GimpPixelRgn rgn_out_nl;
	guchar *rect_out_nl;
	
	if (VERBOSE) printf("== Exporting to a layer..\n");
	
	//creating and attaching new layer
	layer_ID = gimp_layer_new (image_ID, "Cartoon", maindata.width, maindata.height, GIMP_RGBA_IMAGE, 100.0,  GIMP_NORMAL_MODE);
	new_layer = gimp_drawable_get (layer_ID);
	gimp_image_add_layer (image_ID,layer_ID,-1);
	

	rect_out_nl      = g_new (guchar, maindata.width * maindata.height * 4);	
	gimp_pixel_rgn_init (&rgn_out_nl, new_layer, 0, 0, maindata.width, maindata.height, FALSE, TRUE);

	////populating rgn_out with data
	ITERATE(x,0,maindata.width) {ITERATE (y,0,maindata.height) {
		basepos_ch=basepos(x,y,maindata.channels,maindata.width);
		basepos_4=basepos(x,y,4,maindata.width);
		rect_out_nl[basepos_4    ] = rect_out[basepos_ch ];
		rect_out_nl[basepos_4 + 1] = rect_out[basepos_ch+1];
		rect_out_nl[basepos_4 + 2] = rect_out[basepos_ch+2];
		if ( maindata.channels ==4)
			rect_out_nl[basepos_4 + 3] = rect_out[basepos_ch + 3 ];	
		else
			rect_out_nl[basepos_4 + 3] =255;
		}}
   
	//nahratie udajov do rgn_out
    gimp_pixel_rgn_set_rect (&rgn_out_nl, rect_out_nl,0, 0,maindata.width,maindata.height);
	
	gimp_drawable_flush (new_layer);
	gimp_drawable_merge_shadow (new_layer->drawable_id, TRUE);
	gimp_drawable_update (new_layer->drawable_id,0, 0, maindata.width,maindata.height);
	
	gimp_progress_end();
	
	g_free (rect_in) ;
	g_free (rect_out);
	g_free (rect_good);
	g_free (r1);g_free (g1);g_free (b1);g_free (r2);g_free (g2);g_free (b2);
	g_free (br1);g_free (br2);
	g_free (brorig);
	g_free (br_unch);
}	

void cb_changed(GtkToggleButton *cleanbutton){
	maindata.doclean=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (cleanbutton));
	if (DEBUG) printf("clean button changed %d\n",gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (cleanbutton)) );
	}
	
void  mode_changed( GtkComboBox *combo) {
    
    if ( strncmp( gtk_combo_box_get_active_text( combo ), mode0,20) == 0) {
    	maindata.mode=NOCHANGE;}
    if ( strncmp((gtk_combo_box_get_active_text( combo )),mode1,20) == 0) {
    	maindata.mode=CARTREDBR;}
    if ( strncmp((gtk_combo_box_get_active_text( combo )),mode2,20) == 0) {
    	maindata.mode=CARTOONIZE;}
    if ( strncmp((gtk_combo_box_get_active_text( combo )),mode3,10) == 0) {
    	maindata.mode=DOLINES;}
    	
    if (VERBOSE) printf ("Selected color mode: %.d\n",maindata.mode);
    	}


inline gint basepos(gint x, gint y,gint ch, gint width){
	return ch*y*width + ch*x;}

	
inline gfloat get_bright (gfloat R,gfloat G,gfloat B){
	gfloat avg;
	
	avg= 0.299*R + 0.587*G + 0.114*B;
	//avg=(R+G+B)/3.0;
    return avg;}

void exportwrapper (GimpDrawable *drawable){ 
	maindata.exportlayer=TRUE;
	process (drawable,NULL);
	maindata.exportlayer=FALSE;	}

void *create_lines (void *arg) {
	threaddatastr *data;
	gint startheight,endheight,c, locx,locy, x, y;
	gint basepos1;
	gint basepos_local;
	gfloat R,G,B,R_local,B_local,G_local,color_diff,br_diff;
	gboolean isline;
	//gfloat finalvalue;

	data=(threaddatastr *) arg;
	
	startheight=data->startline;
	endheight=data->endline;

	if (DEBUG ) printf (" *create_lines()   thread initiated, range: %.d - %.d. (tid: %.d)\n",\
	startheight,endheight,data->id);
	
	//iteration over central pixels
	ITERATE(y,startheight,endheight) { ITERATE(x,0,maindata.width) {
		basepos1=basepos(x,y,1,maindata.width);
		R=r1[basepos1];
		G=g1[basepos1];		
		B=b1[basepos1];		

		//iterate over near pixels
		isline=FALSE;
		for (c=0;c<8;c=c+1) {
			locx=x+near_pixels[c*2];
			locy=y+near_pixels[c*2+1];
			if (locx <0 || locy <0 || locx>=maindata.width || locy >= maindata.height ) continue;
			basepos_local=basepos1 - basepos_diff[c];
			R_local=r1[basepos_local];
			G_local=g1[basepos_local];			
			B_local=b1[basepos_local];
			color_diff=(R_local - R)*(R_local - R) + (G_local - G)*(G_local - G) +
				(B_local - B)*(B_local - B);
			br_diff=ABS(br1[basepos_local] - br1[basepos1]);
			if (color_diff>(maindata.treshsq/maindata.linesens) || br_diff > (maindata.brtresh/maindata.linesens)) {
				isline=TRUE;break;}
			} // end of iteration over near pixels

		if (isline) br2[basepos1]=0;
		else br2[basepos1]=1;

		}}
	return 0;	
}

void *blur_step (void *arg) {
	threaddatastr *data;
	gint basepos1,localpos1;
	gint x,y,c,d,goodcount;
	gfloat br_diff,hue_diff;
	gfloat r_central,b_central,g_central,br_central;
	gfloat newr,newb,newg,newbr;
	gint maxbasepos1;         // no need to calculate the same number again and again
	gint startheight,endheight;
	gboolean debug=FALSE;
	gint localcount;
	gint multi; //multiplicator for good pixels
	dirstruct dir[8];

	
	data=(threaddatastr *) arg;

	startheight=data->startline;
	endheight=data->endline;

	maxbasepos1=basepos(maindata.width-1,maindata.height-1,1,maindata.width);
	
	ITERATE(y,startheight,endheight) { ITERATE(x,0,maindata.width) {
		basepos1	=basepos(x,y,1,	maindata.width);
		r_central	=r_orig[basepos1];
		g_central	=g_orig[basepos1];
		b_central	=b_orig[basepos1];
		br_central	=brorig[basepos1];
		
		if (x%100==0&& data->id==1  && debug){
		if (isnan(data->rsource[basepos1]) || isnan(data->gsource[basepos1]) || isnan(data->bsource[basepos1]) || isnan(data->brsource[basepos1]) ){
			printf ("sources are nan (r,g,b,br): %.3f %.3f %.3f %.3f, goodcount: %1d\n\nexiting...\n\n",
			data->rsource[basepos1], data->gsource[basepos1] , data->bsource[basepos1], data->brsource[basepos1],goodcount);
			exit(1);}}

		
		//testing directions
		ITERATE(c,0,8){
			dir[c].good=TRUE;
			dir[c].sumbrdist=0;
			dir[c].sumhuedist=0;
			dir[c].sumr=0; dir[c].sumg=0; dir[c].sumb=0;
			dir[c].sumbr=0;
			dir[c].len_good=0;
			ITERATE(d,1,maindata.radius+1){
				localpos1=basepos1+d*maindata.directions[c];				
				if (localpos1 < 0 || localpos1>maxbasepos1){//we are out of image
					if 	(dir[c].len_good==0 )
					dir[c].good=FALSE; 
					break;}
				br_diff=ABS(br_central-data->brsource[localpos1]);
				if (br_diff > maindata.brtresh) {
					dir[c].good=FALSE;
					break;}
				hue_diff=	SQ((r_central - data->rsource[localpos1])) +
							SQ((g_central - data->gsource[localpos1])) +
							SQ((b_central - data->bsource[localpos1]));				
				if (hue_diff>maindata.treshsq) {
					dir[c].good=FALSE;
					break;}
				//so the pixel is OK and can be added to statistics for this direction
				dir[c].sumr+=data->rsource[localpos1];	
				dir[c].sumg+=data->gsource[localpos1];
				dir[c].sumb+=data->bsource[localpos1];		
				dir[c].sumbr+=data->brsource[localpos1];
				dir[c].len_good+=1;	
			}
		}
		
		//iterating over results:
		newr=0;newg=0; newb=0;newbr=0;goodcount=0;
		ITERATE(c,0,8){
			if(dir[c].good) {
				newr+=dir[c].sumr/dir[c].len_good;
				newg+=dir[c].sumg/dir[c].len_good;			
				newb+=dir[c].sumb/dir[c].len_good;
				newbr+=dir[c].sumbr/dir[c].len_good;	
				goodcount+=1;}
			}
		if (x%100==0&& data->id==1 && debug) printf (" Goodcount: %1d\n", goodcount);
		
		//insterting statistics
		if (goodcount<=maindata.mindircount) data->stat[0]+=1;
		else {
			data->stat[1]+=1;
			data->stat[2]+=goodcount;}
			
		//PROCESSING
		if (goodcount<=maindata.mindircount && maindata.doclean==FALSE) {
			data->rtarget[basepos1]=data->rsource[basepos1] ;
			data->gtarget[basepos1] =data->gsource[basepos1] ;
			data->btarget[basepos1] =data->bsource[basepos1] ;
			data->brtarget[basepos1]=data->brsource[basepos1] ;
			continue;}

		if (goodcount>maindata.mindircount && maindata.doclean==FALSE) {
			data->rtarget[basepos1] =newr/goodcount ;
			data->gtarget[basepos1] =newg/goodcount ;
			data->btarget[basepos1] =newb/goodcount ;
			data->brtarget[basepos1]=newbr/goodcount ;
			continue;}			
						
		// we need to get newX (as because of goodcount<=maindata.mindircount none were calculated)
		if (goodcount<=maindata.mindircount) {
			newr=0;newg=0; newb=0;newbr=0;localcount=0;
			ITERATE(c,0,8) {
				localpos1=basepos1+maindata.directions[c];
				if (localpos1 < 0 || localpos1>maxbasepos1) continue;
				if (rect_good[localpos1]==TRUE) multi=15;
				else multi=1;
				//if (x%100==0&& data->id==1 && debug) printf ("   multi: %1d\n",multi);
				newr+=data->rsource[localpos1]*multi;
				newg+=data->gsource[localpos1]*multi;
				newb+=data->bsource[localpos1]*multi;
				newbr+=data->brsource[localpos1]*multi;
				if (x%100==0&& data->id==1 && debug) printf ("   multi: %2d, Br: %.3f\n",multi,data->brsource[localpos1]);
				localcount+=multi;
				}	
			newr=newr/localcount;
			newg=newg/localcount;
			newb=newb/localcount;
			newbr=newbr/localcount;}
		else{
			newr=newr/goodcount;
			newg=newg/goodcount;
			newb=newb/goodcount;
			newbr=newbr/goodcount;}			

		//setting good/bad pixel flag
		if (goodcount<=maindata.mindircount) rect_good[basepos1]=FALSE;
		else rect_good[basepos1]=TRUE;
		
		if (x%100==0&& data->id==1 && debug)  printf ("  smoothening values: %.3f %.3f %.3f %.3f\n",
		newr,newg,newb,newbr);
		
		//the rest of cycle is for forced smoothening
		data->rtarget[basepos1] =newr;
		data->gtarget[basepos1] =newg;
		data->btarget[basepos1] =newb;
		data->brtarget[basepos1]=newbr;

		r_orig[basepos1] =0.7*r_orig[basepos1] +0.3*newr;	
		g_orig[basepos1] =0.7*g_orig[basepos1] +0.3*newg;	
		b_orig[basepos1] =0.7*b_orig[basepos1] +0.3*newb;	
		brorig[basepos1] =0.7*brorig[basepos1] +0.3*newbr;
	}}
		
	return 0;
}

void blur_data() {
	//this is controlling function to initiate single blurring steps
	gint c,d,e;
	gint resultlayer; // where final data ends
	gint stats[3];
	gfloat statsfloat[3];

	if (DEBUG) printf("  Blurring iterations to do: %.d\n",maindata.iterations);
	resultlayer=1;
	
	//populating thread data
	ITERATE(c,0,THREADS) {
		if (c == 0 ) thread_data[c].startline=0;
		else thread_data[c].startline=thread_data[c-1].endline;
		if ((c+1) == THREADS ) thread_data[c].endline=maindata.height;
		else thread_data[c].endline=maindata.height * (c+1) / THREADS; 
		}
		
	//ITERATE(c,0,THREADS) printf ("startline = %.d, endline= %.d\n",thread_data[c].startline, thread_data[c].endline);
		
	ITERATE(c,0,maindata.iterations) {
		if (DEBUG ) printf ("  Initiating iteration: %1d\n", c+1);
		//calculating some values for this blur iteration
		ITERATE(d,0,THREADS) {
			thread_data[d].id=d;
			//resetting statistics
			thread_data[d].stat[0]=0;
			thread_data[d].stat[1]=0;
			thread_data[d].stat[2]=0;			
			// definyng source and target arrays
			if (c%2 == 0) {
				thread_data[d].rsource=r1;
				thread_data[d].gsource=g1;
				thread_data[d].bsource=b1;
				thread_data[d].brsource=br1;				
				thread_data[d].rtarget=r2;
				thread_data[d].gtarget=g2;
				thread_data[d].btarget=b2;
				thread_data[d].brtarget=br2;
				resultlayer=2;}
			else {
				thread_data[d].rsource=r2;
				thread_data[d].gsource=g2;
				thread_data[d].bsource=b2;
				thread_data[d].brsource=br2;				
				thread_data[d].rtarget=r1;
				thread_data[d].gtarget=g1;
				thread_data[d].btarget=b1;
				thread_data[d].brtarget=br1;
				resultlayer=1;}				
			
			thread_data[d].iteration=c;
				
			//starting thread
			gth[d]=g_thread_create( (GThreadFunc) blur_step,&thread_data[d],TRUE,NULL);
			//gth[d]=g_thread_new( "blurring"c, (GThreadFunc) blur_step,&thread_data[d]);			
			//if (DEBUG) sleep(1);
			 }
		
		ITERATE(d,0,THREADS) { //waiting till all of them ends
			g_thread_join(gth[d]);    }

		//collecting statistics
		stats[0]=0;stats[1]=0;stats[2]=0;
		ITERATE(e,0,THREADS){
			stats[0]+=thread_data[e].stat[0];
			stats[1]+=thread_data[e].stat[1];
			stats[2]+=thread_data[e].stat[2];}			
		statsfloat[0]=(gfloat)stats[0]*100/(stats[0]+stats[1]);
		statsfloat[1]=(gfloat)stats[1]*100/(stats[0]+stats[1]);
		statsfloat[2]=(gfloat)stats[2]*100/(maindata.height*maindata.width*8);
						
		//blur_step(width,height,maindata.endlayer, threshred);
		
		if (VERBOSE){
			printf ("  Update progress: %.3f\n",0.1 + (double)(((c+1)/(double)maindata.iterations)*0.8) );
			printf ("   Statistics: Not blurrable: %2.2f%%, blurrable:  %2.2f%%. overall  %2.2f%%\n",statsfloat[0],statsfloat[1],statsfloat[2]);}
		if (maindata.preview==FALSE){ gimp_progress_update( 0.1 + (double)(((c+1)/(double)maindata.iterations)*0.8));}

		} // end of single iteration steps
		
		//the final data should end up in tmp1 and br1
	if (resultlayer ==2) {
		if (DEBUG) printf ("Copying data back to tmp1&br1\n");
		ITERATE(c,0,maindata.width * maindata.height){
			br1[c]=br2[c]; 
			r1[c]=r2[c];
			g1[c]=g2[c];
			b1[c]=b2[c];}}
}

void do_lines() {
	//doing cartoon lines
	gint d,x,y,basepos_1,changedcount,samecount;
	changedcount=0;
	samecount=0;
	
	if (DEBUG ) printf (" Initiating cartoon lines creation.\n");			
	
	ITERATE(d,0,THREADS) {
		gth[d]=g_thread_create( (GThreadFunc) create_lines,&thread_data[d],TRUE,NULL); }

	ITERATE(d,0,THREADS) { //waiting till all of them ends
		g_thread_join(gth[d]);    }

	//cleaning and copying back to br1 (line=0)
	ITERATE(y,0,maindata.height) { ITERATE(x,0,maindata.width) {
		basepos_1=basepos(x,y,1,maindata.width); 
		if (  x==0 || y==0 || x==maindata.width-1 || y==maindata.height-1 ){
			br1[basepos_1]=br2[basepos_1];
			continue;}
		samecount=0;
		ITERATE(d,0,8) {
			if (br2[basepos_1+maindata.directions[d]]==br2[basepos_1]) {
				samecount+=1;
				if (samecount>2) break;}}
		if (samecount >2){ br1[basepos_1]=br2[basepos_1];}
		else {br1[basepos_1] = ((br2[basepos_1]==1)? 0 : 1);
			changedcount+=1;}
		}}					
	if (VERBOSE) printf (" Cleaning lines - changed pixels: %1d\n", changedcount);
	}

void restore_bright(){
	gint y,x,basepos_1;
	
	ITERATE(y,0,maindata.height) { ITERATE(x,0,maindata.width) {
		basepos_1=basepos(x,y,1,maindata.width);
		br1[basepos_1] = br_unch[basepos_1];	
		}}
	}	

void shift_bright(gfloat start, gfloat end){
	gint y,x,basepos_1;
	ITERATE(y,0,maindata.height) { ITERATE(x,0,maindata.width) {
	basepos_1=basepos(x,y,1,maindata.width);
	br1[basepos_1] = start + br1[basepos_1] * (end-start);}}
	}

void findminmax(gfloat v1, gfloat v2,gfloat v3,gfloat* min, gfloat* max){
	*min=v1;
	if (v2<*min) *min=v2;	
	if (v3<*min) *min=v3;	
	*max=v1;
	if (v2>*max) *max=v2;	
	if (v3>*max) *max=v3;	}

		
void change_saturation(){
	gint y,x,basepos_1;
	gfloat sat,newsat,boost;
	gboolean debug=FALSE;
	
	if (maindata.slidersat ==1) return;

	
	ITERATE(y,0,maindata.height) { ITERATE(x,0,maindata.width) {
		////basepos_out=basepos(x,y,maindata.channels,maindata.width);
		basepos_1=basepos(x,y,1,maindata.width);	
		
		////if (debug && x==100 && y%10==0) printf(" rgb: %.3f %.3f %.3f\n",r1[basepos_1],g1[basepos_1],b1[basepos_1]);
		if (r1[basepos_1]==g1[basepos_1] && r1[basepos_1]==b1[basepos_1]) continue;
		
		//avg=br1[basepos_1];
		sat=(ABS(r1[basepos_1]) + ABS(r1[basepos_1]) + ABS(r1[basepos_1]))/3.0;

		if (maindata.slidersat >1) {
			if(sat>maindata.breakpoint) newsat=sat;
			else newsat=pow(sat/maindata.breakpoint,1/maindata.slidersat)*maindata.breakpoint;
			if (newsat-sat>0.05) newsat=sat+0.05;
 			boost=MIN(newsat/sat,5.0);
			if (debug && y%10==0 && x%10==0 ) printf (" Old sat: %.3f, breakpoint: %.3f , rel boost: %.3f, abs increase: %.3f\n",
				sat,maindata.breakpoint,boost,sat*boost-sat);
			}
		//following two options are for slidersat <1
		else {
			if (maindata.slidersat*0.15<sat) boost=maindata.slidersat*0.15/sat;
			else boost=1;
			if (debug && y%10==0 && x%10==0 ) printf ( "sat: %.3f, slider: %.3f, boost: %.3f\n",
				sat, maindata.slidersat*0.15,boost);
			}
		
		//boost=MIN(sat/maindata.breakpoint,maindata.slidersat/maindata.breakpoint);


		//changing values
		r1[basepos_1]=r1[basepos_1]*boost;
		g1[basepos_1]=g1[basepos_1]*boost;
		b1[basepos_1]=b1[basepos_1]*boost;	
		
		//???
		////shifting brightness to have RGB within 0-255 range (might not be possible)
		//avg=br1[basepos_1];
		//findminmax(r1[basepos_1],g1[basepos_1],b1[basepos_1],&minRGB,&maxRGB);
		//maxRGB=maxRGB+avg;
		//minRGB=minRGB+avg;
		
		////if ((maxRGB-minRGB) <=1) {
			////if (minRGB<0) br1[basepos_1]=br1[basepos_1]-minRGB;
			////if (maxRGB>1) br1[basepos_1]=br1[basepos_1]-maxRGB+1;}
		}}	
	}
				
					
void fill_rect_out() {
	gint x,y;
	gfloat r,g,b;
	gint basepos_1, basepos_out;
	gfloat bright;

	ITERATE(y,0,maindata.height) { ITERATE(x,0,maindata.width) {
		basepos_out=basepos(x,y,maindata.channels,maindata.width);
		basepos_1=basepos(x,y,1,maindata.width);

		if (maindata.mode == DOLINES) {
			rect_out[basepos_out  ]=TO_INT_W_GAMMA(br1[basepos_1]);
			rect_out[basepos_out+1]=rect_out[basepos_out  ];
			rect_out[basepos_out+2]=rect_out[basepos_out  ];
			continue;}
		else 
			bright = br1[basepos_1];

		r=r1[basepos_1]+bright;
		g=g1[basepos_1]+bright;
		b=b1[basepos_1]+bright;
		if (r<0) r=0;if (g<0) g=0;if (b<0) b=0;
		if (r>1) r=1;if (g>1) g=1;if (b>1) b=1;
		rect_out[basepos_out  ]=TO_INT_W_GAMMA(r);
		rect_out[basepos_out+1]=TO_INT_W_GAMMA(g);
		rect_out[basepos_out+2]=TO_INT_W_GAMMA(b);			
		if (maindata.channels==4) rect_out[basepos_out+3]=rect_in[basepos_out+3];
		}}
}


void fill_tmp() {
	gint x,y;
	gint basepos_in,basepos_1;
	gfloat r,g,b;

	ITERATE(y,0,maindata.height) { ITERATE(x,0,maindata.width) {
		basepos_in  = basepos(x,y,maindata.channels,maindata.width);
		basepos_1   = basepos(x,y,1,maindata.width);

		r=TO_FLOAT_W_GAMMA(rect_in[basepos_in  ]);
		g=TO_FLOAT_W_GAMMA(rect_in[basepos_in+1]);
		b=TO_FLOAT_W_GAMMA(rect_in[basepos_in+2]);		

		br1[basepos_1   ]   = get_bright(r,g,b);
		brorig[basepos_1]   = br1[basepos_1];
		r1[basepos_1]		= r-br1[basepos_1];
		g1[basepos_1]		= g-br1[basepos_1];
		b1[basepos_1]		= b-br1[basepos_1];
		r_orig[basepos_1]  	= r1[basepos_1];
		g_orig[basepos_1]  	= g1[basepos_1];
		b_orig[basepos_1]  	= b1[basepos_1];
		if (maindata.mode==NOCHANGE) br_unch[basepos_1] = br1[basepos_1];
		}}
}	

void process (GimpDrawable *source,GimpPreview *preview) {

	gint         x1, y1, x2, y2; 		//x,y from image not area
	GimpPixelRgn rgn_in, rgn_out;
	gint i,c;
	gint basepos_c,basepos_l,locx,locy;

	if (VERBOSE ) {
		printf ("== Running Selective Color Blur plugin... ==\n");
		} // check this

 	//make sure g_thread is running
 	if( !g_thread_supported() )  g_thread_init(NULL);
 	if( !g_thread_supported() ) {
     	printf("g_thread NOT supported\n");
     	exit (1);}

	// getting coordinates of what will be saturated
	if (preview){
		gimp_preview_get_position (preview, &x1, &y1);
		gimp_preview_get_size (preview, &maindata.width, &maindata.height);
		x2 = x1 + maindata.width;
		y2 = y1 + maindata.height;
		maindata.preview=TRUE;
		}
    else {
		gimp_drawable_mask_bounds (source->drawable_id,&x1, &y1,&x2, &y2);
		maindata.width = x2 - x1;
		maindata.height = y2 - y1;
		maindata.preview=FALSE;}
     
     if (!maindata.preview) gimp_progress_init ("Selective Color Blur...");
     
	 //converting tresh from raw to square
	 maindata.treshsq=pow(maindata.tresh,2);
	 maindata.shortrad=(gint)maindata.radius*0.7; // ? needed
	    
	 maindata.channels = gimp_drawable_bpp (source->drawable_id);//getting number of channels

	for (i=0;i<256;i++) { //to speed up conversion from raw RGB to gammed values
		gammed_values[i]=pow((gfloat)i/255,1/2.2);}

  //initializing pixelrgn for input and output row
  gimp_pixel_rgn_init (&rgn_in,
                       source,
                       x1, y1,
                       maindata.width, maindata.height,
                       FALSE, FALSE);
                       
  gimp_pixel_rgn_init (&rgn_out,
                       source,
                       x1, y1,
                       maindata.width, maindata.height,
                       preview == NULL, TRUE);
                       
    if (maindata.width ==1) return;

	if(maindata.slidersat >1){
		if(maindata.slidersat >1.5)  maindata.breakpoint=0.15+(maindata.slidersat-1)*0.10;
		else  maindata.breakpoint=0.15;}
	else maindata.breakpoint=0.25;

	if (VERBOSE) {
		printf (" Processed area: %1d x %1d, channels: %1d\n",maindata.width,maindata.height,maindata.channels);
		printf (" Processing mode: %1d\n",maindata.mode);
		printf (" Radius: %1d px,  Iterations: %1d\n",	maindata.radius,maindata.iterations);
		printf (" Hue tresh: %.3f / %.4f, brtresh: %.3f\n",maindata.tresh, maindata.treshsq,  maindata.brtresh);
		printf (" Lines creation sensitivity: %.3f %s\n", maindata.linesens,(maindata.mode!=3)?"(not used now)":"");
		printf (" Brightness modification: %.2f - %.2f\n", maindata.darkpoint,maindata.lightpoint);
		printf (" Slider saturation: %.2f, breakpoint: %.3f \n",maindata.slidersat, maindata.breakpoint);
		printf (" Minimum good directions: %1d\n", maindata.mindircount);		
		printf (" Small details cleaning: %s\n",(maindata.doclean)?"yes":"no");
		printf (" Exporting as new layer: %s\n", (maindata.exportlayer)?"yes":"no");
		}
	
 	/* Initialise memory for rects*/
	rect_in  =   g_new (guchar, maindata.channels  * maindata.width * maindata.height); 
	rect_out= g_new (guchar, maindata.channels * maindata.width * maindata.height);
	rect_good=g_new (gboolean,maindata.width * maindata.height);
	r1	=g_new (gfloat, maindata.width * maindata.height ); 
	g1	=g_new (gfloat, maindata.width * maindata.height ); 
	b1	=g_new (gfloat, maindata.width * maindata.height ); 
	r2	=g_new (gfloat, maindata.width * maindata.height ); 
	g2	=g_new (gfloat, maindata.width * maindata.height ); 
	b2	=g_new (gfloat, maindata.width * maindata.height ); 
	br1	=g_new (gfloat, maindata.width * maindata.height);
	br2	=g_new (gfloat, maindata.width * maindata.height);
	r_orig=g_new (gfloat, maindata.width * maindata.height);
	g_orig=g_new (gfloat, maindata.width * maindata.height);
	b_orig=g_new (gfloat, maindata.width * maindata.height);	
	brorig= g_new (gfloat, maindata.width * maindata.height);
	if (maindata.mode==NOCHANGE) br_unch=g_new (gfloat, maindata.width * maindata.height);

	//marking all pixels unsafe
	if (maindata.doclean) ITERATE(c,0,maindata.width*maindata.height) {rect_good[c]=FALSE;};

	if (DEBUG) printf (" CB: memory allocated\n");
	
	if (DEBUG && maindata.preview) printf (" Doing preview\n");
	if (!maindata.preview) gimp_progress_init ("Selective color blur...");
	
	//calculating direction differentials 
	maindata.directions[0]=basepos(1,2,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[1]=basepos(2,1,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[2]=basepos(1,0,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[3]=basepos(0,1,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[4]=basepos(2,2,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[5]=basepos(2,0,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[6]=basepos(0,0,1,maindata.width) - basepos(1,1,1,maindata.width);
	maindata.directions[7]=basepos(0,2,1,maindata.width) - basepos(1,1,1,maindata.width);
	if (DEBUG) printf (" Directions stepping is: %.d, %.d, %.d and %.d\n", maindata.directions[0],maindata.directions[1],\
	maindata.directions[2],maindata.directions[3]);

	//calculating basepos_diff
	if (maindata.mode==DOLINES || maindata.doclean) {
		basepos_c=basepos(1,1,1,maindata.width);
		for (c=0;c<8;c=c+1) {
			locx=1+near_pixels[c*2];
			locy=1+near_pixels[c*2+1];
			basepos_l     =basepos(locx,locy,1,maindata.width);
			basepos_diff[c]   =basepos_c - basepos_l;
			if (DEBUG) printf ("  DEBUG: Basepos diff: %.d, (%.d x %.d)\n",basepos_diff[c],locx,locy);
			}
		}
	
	//loading all data to rect_in
	gimp_pixel_rgn_get_rect (&rgn_in,rect_in,x1, y1,maindata.width,maindata.height);
	if (DEBUG) printf (" CB: data loaded in rect_in\n");

	//fill input data to rect_tmp
	fill_tmp();
	if (DEBUG) printf (" CB: primary data loaded to rect_tmp\n");
	
	if ( maindata.preview == FALSE ) gimp_progress_update (0.1);
	
	//blurring data
	blur_data ();
	if (DEBUG) printf (" CB: blurring done\n");
	
	if (maindata.mode == DOLINES) 
		do_lines();
	else {
		if (maindata.mode==NOCHANGE)	restore_bright();
		if (maindata.slidersat != 0)	change_saturation();
		if (!(maindata.darkpoint==0 && maindata.lightpoint==1)) shift_bright(maindata.darkpoint,maindata.lightpoint);
		else if (maindata.mode==CARTREDBR) shift_bright(0.3,0.9);
		}
		
		
	//calculating data
	fill_rect_out();
	if (DEBUG) printf (" CB: rect_out filled\n");
	if (DEBUG) printf( "DEBUG: rect_out [0,0]: %1d, %1d, %1d\n", rect_out[0],rect_out[1],rect_out[1]);

	if (maindata.exportlayer) {
		export_to_layer();
		gimp_progress_end();
		return;}

	//putting data back to drawable
    gimp_pixel_rgn_set_rect (&rgn_out, rect_out,x1, y1,maindata.width,maindata.height);
 	if (DEBUG) printf ("CB: data back in rgn\n");   

	g_free (rect_in) ;
	g_free (rect_out);
	g_free (rect_good);
	g_free (r1);g_free (g1);g_free (b1);g_free (r2);g_free (g2);g_free (b2);
	g_free (r_orig),g_free(g_orig), g_free(b_orig);
	g_free (br1);g_free (br2);
	g_free (brorig);
	if (maindata.mode==NOCHANGE) g_free (br_unch);
	if (DEBUG) printf ("rect freeied\n");  

  //update (graphical) of preview/image
	if (preview) {
    	gimp_drawable_preview_draw_region (GIMP_DRAWABLE_PREVIEW (preview),
        &rgn_out);}

    else {
		gimp_drawable_flush (source);
		gimp_drawable_merge_shadow (source->drawable_id, TRUE);
		gimp_drawable_update (source->drawable_id,x1,y1,maindata.width,maindata.height);}
	if (DEBUG) printf ("drawable updated\n");
	gimp_progress_end();
}
 

gboolean plugin_gui (GimpDrawable *source)
{
  GtkWidget *dialog;
  GtkWidget *main_vbox,*table,*sat_hbox;
  GtkWidget *label_radius, *label_tresh, *label_iter, *br_tresh,*lines_sens,*label_sat;
  GtkWidget *preview;
  GtkObject *radius_spin_adj, *iter_spin_adj, *tresh_spin_adj, *brtresh_spin_adj,*linesens_spin_adj,*slider_sat_adj;
  GtkWidget *exportbutton;
  GtkWidget *eb_align,*t_align,*m_align,*s_align;

  gimp_ui_init ("Color Blur", FALSE);
  
  gimp_dialogs_show_help_button (FALSE);
  dialog = gimp_dialog_new (title, "colorblur",
                            NULL, 0,
                            gimp_standard_help_func, "plug-in-colorblur",
                            GIMP_STOCK_RESET, RESPONSE_RESET,
                            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                            GTK_STOCK_OK,     RESPONSE_OK,
                            NULL);

  
	gtk_container_set_border_width (GTK_CONTAINER (dialog), 12); // ???

  
	main_vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), main_vbox);
	gtk_widget_show (main_vbox);
	
	preview = gimp_drawable_preview_new (source, &previewbool);
	gtk_box_pack_start (GTK_BOX (main_vbox), preview, TRUE, TRUE, 0);
	gtk_widget_set_tooltip_text(preview, "Preview is AUTOEXPANDING. Technical comment: \
 When calculating this preview, only visible area is considered. \
 That might lead to slightly different resuls when computing final image afterwards.");
	gtk_widget_show (preview);

	m_align=gtk_alignment_new(0.5,0.5,0.5,0);
	gtk_box_pack_start (GTK_BOX (main_vbox), m_align, FALSE, FALSE, 0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(m_align),15,5,0,0);
	gtk_widget_show (m_align);	
	
	combo = gtk_combo_box_new_text();
	gtk_combo_box_append_text( GTK_COMBO_BOX( combo),mode0);
	gtk_combo_box_append_text( GTK_COMBO_BOX( combo),mode1);
	gtk_combo_box_append_text( GTK_COMBO_BOX( combo),mode2);
	gtk_combo_box_append_text( GTK_COMBO_BOX( combo),mode3);
	gtk_container_add(GTK_CONTAINER(m_align), combo);
	gtk_widget_set_tooltip_text(combo,"Select what to do");	
	gtk_combo_box_set_active(GTK_COMBO_BOX( combo), MODEDEF);  //default
	gtk_widget_show (combo);
	
	//creating table 
	t_align=gtk_alignment_new(0,0.5,0,0);
	gtk_box_pack_start (GTK_BOX (main_vbox), t_align, FALSE, FALSE, 0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(t_align),25,15,10,0);
	gtk_widget_show (t_align);	
	
	table=gtk_table_new(2,5,FALSE);
	gtk_table_set_col_spacings( GTK_TABLE(table), 25 );
	gtk_table_set_row_spacings( GTK_TABLE(table), 3 );
	gtk_container_add(GTK_CONTAINER(t_align), table);

		//inserting labels
		label_radius = gtk_label_new ("<b>Blur Radius:</b>");
		gtk_label_set_use_markup (GTK_LABEL(label_radius), TRUE);
		gtk_misc_set_alignment (GTK_MISC (label_radius), 0, 0.5);
		gtk_table_attach_defaults(GTK_TABLE(table),label_radius,0,1,0,1);
		gtk_widget_show (label_radius);
		label_iter = gtk_label_new ("<b>Blur Iterations:</b>");
		gtk_label_set_use_markup (GTK_LABEL(label_iter), TRUE);
		gtk_misc_set_alignment (GTK_MISC (label_iter), 0, 0.5);
		gtk_table_attach_defaults(GTK_TABLE(table),label_iter,0,1,1,2);
		gtk_widget_show (label_iter);
		label_tresh = gtk_label_new ("<b>Color Treshold:</b>");
		gtk_label_set_use_markup (GTK_LABEL(label_tresh), TRUE);
		gtk_misc_set_alignment (GTK_MISC (label_tresh), 0, 0.5);
		gtk_table_attach_defaults(GTK_TABLE(table),label_tresh,0,1,2,3);
		gtk_widget_show (label_tresh);
		br_tresh = gtk_label_new ("<b>Value Treshold:</b>");
		gtk_label_set_use_markup (GTK_LABEL(br_tresh), TRUE);
		gtk_misc_set_alignment (GTK_MISC (br_tresh), 0, 0.5);
		gtk_table_attach_defaults(GTK_TABLE(table),br_tresh,0,1,3,4);
		gtk_widget_show (br_tresh);
		lines_sens = gtk_label_new ("<b>Lines sensitivity:</b>");
		gtk_label_set_use_markup (GTK_LABEL(lines_sens), TRUE);
		gtk_misc_set_alignment (GTK_MISC (lines_sens), 0, 0.5);
		gtk_table_attach_defaults(GTK_TABLE(table),lines_sens,0,1,4,5);
		gtk_widget_show (lines_sens);


  		//inserting spinboxes
		radius_spin_adj = gtk_adjustment_new (RADIUSDEFAULT, 0, 100, 1, 10,0);
		radius_spin = gtk_spin_button_new (GTK_ADJUSTMENT (radius_spin_adj), 2, 0);
		gtk_table_attach_defaults(GTK_TABLE(table),radius_spin,1,2,0,1);
		gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (radius_spin), TRUE);
		gtk_widget_set_tooltip_text( radius_spin, "Radius per iteration, use smaller radius for more fragmented images or simply for smaller images.");
		gtk_widget_show (radius_spin);	
		iter_spin_adj = gtk_adjustment_new (ITERDEFAULT, 0, 100, 1, 10,0);
		iter_spin = gtk_spin_button_new (GTK_ADJUSTMENT (iter_spin_adj), 2, 0);
		gtk_table_attach_defaults(GTK_TABLE(table),iter_spin,1,2,1,2);
		gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (iter_spin), TRUE);
		gtk_widget_set_tooltip_text( iter_spin, "Use higher numbers (even above like 30) if working in Forced Smoothing mode.");
		gtk_widget_show (iter_spin);
		tresh_spin_adj = gtk_adjustment_new (TRESHDEFAULT,  0, 2, 0.005, 0.1,0);
		tresh_spin = gtk_spin_button_new (GTK_ADJUSTMENT (tresh_spin_adj),0.001, 3);
		gtk_table_attach_defaults(GTK_TABLE(table),tresh_spin,1,2,2,3);
		gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (tresh_spin), TRUE);
		gtk_widget_set_tooltip_text( tresh_spin, "3D distance in normalized RGB colorspace, brightness is not considered here.");
		gtk_widget_show (tresh_spin);	
		brtresh_spin_adj = gtk_adjustment_new (BRTRESHDEF,  0, 1, 0.01, 0.1,0);
		brtresh_spin = gtk_spin_button_new (GTK_ADJUSTMENT (brtresh_spin_adj),0.01, 3);
		gtk_table_attach_defaults(GTK_TABLE(table),brtresh_spin,1,2,3,4);
		gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (brtresh_spin), TRUE);
		gtk_widget_set_tooltip_text( brtresh_spin, "Treshold for brightness, set 1 to ignore the brightness during blurring");
		//gtk_widget_set_size_request(tresh_spin,40,25);
		gtk_widget_show (brtresh_spin);	
		linesens_spin_adj = gtk_adjustment_new (LINESENSDEF,  0.1, 20, 0.1, 0.1,0);
		linesens_spin = gtk_spin_button_new (GTK_ADJUSTMENT (linesens_spin_adj),0.1, 1);
		gtk_table_attach_defaults(GTK_TABLE(table),linesens_spin,1,2,4,5);
		gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (linesens_spin), TRUE);
		gtk_widget_set_tooltip_text( linesens_spin, "Controls how many black pixels (~lines) will be generated. \
 A pixel is made black if the RGB / brightness distance from it to any of adjacent pixels is bigger than treshold.\
 (Used only in \"Create lines\" mode)");
		//gtk_widget_set_size_request(tresh_spin,40,25);
		gtk_widget_show (linesens_spin);	
	
	gtk_widget_show (table);	

	s_align=gtk_alignment_new(0.5,0.5,1,0);
	gtk_box_pack_start (GTK_BOX (main_vbox), s_align, FALSE, FALSE, 0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(s_align),0,25,10,50);
	gtk_widget_show (s_align);	

		sat_hbox = gtk_hbox_new (FALSE, 20);
		gtk_container_add (GTK_CONTAINER(s_align), sat_hbox);
		gtk_widget_show (sat_hbox);

			label_sat = gtk_label_new ("<b>Saturation:</b>");
			gtk_label_set_use_markup (GTK_LABEL(label_sat), TRUE);
			gtk_misc_set_alignment (GTK_MISC (label_sat), 0, 0.5);
			gtk_box_pack_start (GTK_BOX (sat_hbox), label_sat, FALSE, FALSE, 0);
			gtk_widget_show (label_sat);

			slider_sat_adj = gtk_adjustment_new (1.0, 0.0, 2, 0.01, 0.1, 0);
			slider_sat = gtk_hscale_new (GTK_ADJUSTMENT (slider_sat_adj));
			gtk_widget_set_tooltip_text(slider_sat, "Very rough and non-linear saturation tool");
			gtk_scale_set_digits(GTK_SCALE (slider_sat), 2) ;
			//gtk_misc_set_alignment (GTK_MISC (slider_sat), 0, 1); 
			//gtk_range_set_inverted(GTK_RANGE (slider_sat),TRUE) ;
			gtk_widget_set_size_request(slider_sat,200,60);
			gtk_box_pack_start (GTK_BOX (sat_hbox), slider_sat, FALSE, FALSE, 0);
			gtk_widget_show (slider_sat);
	  
	  
	  
	  



	cleanbutton= gtk_check_button_new_with_label("Forced smoothing");
	gtk_widget_set_tooltip_text( cleanbutton, 
		"Remove small discrete areas and/or blurred edges. Sacrifices some information to get compact areas. Increase iterations a bit for better effect.");
	gtk_box_pack_start (GTK_BOX (main_vbox), cleanbutton, FALSE, FALSE, 0);
	gtk_widget_set_size_request(cleanbutton,0,25);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cleanbutton), FALSE);
	gtk_widget_show (cleanbutton);
	
	eb_align=gtk_alignment_new(0.5,0.5,0.5,0);
	gtk_box_pack_start (GTK_BOX (main_vbox), eb_align, FALSE, FALSE, 0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(eb_align),15,10,0,0);
	gtk_widget_show (eb_align);	
	
	exportbutton = gtk_button_new_with_label("Export as new layer");
	gtk_widget_set_tooltip_text(exportbutton, "Plugin window will stay alive with all settings preserved. During export/calculation entire plugin's window is frozen.");
	gtk_container_add(GTK_CONTAINER(eb_align), exportbutton);
	gtk_widget_show (exportbutton);	
		
	g_signal_connect_swapped (preview, "invalidated",  G_CALLBACK (process),source);
	g_signal_connect_swapped (radius_spin_adj   , "value_changed", G_CALLBACK (gimp_preview_invalidate),preview);
	g_signal_connect_swapped (iter_spin_adj     , "value_changed", G_CALLBACK (gimp_preview_invalidate),preview);
	g_signal_connect_swapped (tresh_spin_adj    , "value_changed", G_CALLBACK (gimp_preview_invalidate),preview);
	g_signal_connect_swapped (brtresh_spin_adj  , "value_changed", G_CALLBACK (gimp_preview_invalidate),preview);
	g_signal_connect_swapped (linesens_spin_adj , "value_changed", G_CALLBACK (gimp_preview_invalidate),preview);
 	g_signal_connect_swapped (slider_sat_adj,   "value_changed", G_CALLBACK (gimp_preview_invalidate), preview);
	g_signal_connect_swapped (G_OBJECT(combo   ), "changed"      , G_CALLBACK (gimp_preview_invalidate),preview);
	g_signal_connect_swapped (G_OBJECT(cleanbutton),"toggled"    , G_CALLBACK (gimp_preview_invalidate),preview);

	process (source, GIMP_PREVIEW (preview));

	//update of values in maindata
	g_signal_connect (radius_spin_adj, "value_changed", G_CALLBACK (gimp_int_adjustment_update  ), &maindata.radius);  
	g_signal_connect (iter_spin_adj  , "value_changed", G_CALLBACK (gimp_int_adjustment_update  ), &maindata.iterations);  
	g_signal_connect (tresh_spin_adj , "value_changed", G_CALLBACK (gimp_float_adjustment_update), &maindata.tresh);
	g_signal_connect (brtresh_spin_adj,"value_changed", G_CALLBACK (gimp_float_adjustment_update), &maindata.brtresh); 
	g_signal_connect (linesens_spin_adj,"value_changed", G_CALLBACK (gimp_float_adjustment_update), &maindata.linesens); 
	g_signal_connect (slider_sat_adj, "value_changed", G_CALLBACK (gimp_float_adjustment_update), &maindata.slidersat);  
	g_signal_connect (G_OBJECT( combo    ),  "changed", G_CALLBACK ( mode_changed               ), NULL );
	g_signal_connect (G_OBJECT( cleanbutton ),  "toggled", G_CALLBACK ( cb_changed                 ), NULL );
	g_signal_connect_swapped(G_OBJECT(exportbutton),"clicked",G_CALLBACK( exportwrapper ), source );
	gtk_widget_show (dialog);
  
	//receiving signals from gimp buttons + "x" (close window button)
	g_signal_connect (dialog, "response",
                    G_CALLBACK (response_callback),
                    NULL);

	gtk_main ();
	
	return process_image; //TRUE if image should be processed

}

void response_callback (GtkWidget *widget,  gint response_id) {
	
  switch (response_id) 	 {
	case RESPONSE_RESET:
		if (VERBOSE)  printf ("Resetting... \n");
		
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(radius_spin),RADIUSDEFAULT);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(iter_spin),ITERDEFAULT);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(tresh_spin),TRESHDEFAULT);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(brtresh_spin),BRTRESHDEF);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(linesens_spin),LINESENSDEF);
		gtk_range_set_value(GTK_RANGE(slider_sat),1.0);
		gtk_combo_box_set_active(GTK_COMBO_BOX( combo), MODEDEF);
		break;

    case RESPONSE_OK:  //quitting plugin window and applying change
	  //printf ("OK... \n");
	  process_image=TRUE; 
	  gtk_widget_destroy (widget);
	  gtk_main_quit ();     
	  break;

    default: // if other response - terminate plugin window
      gtk_widget_destroy (widget);
      gtk_main_quit ();
      break;
    };
}

